/*
 * This file is part of firk's Common C Library
 *  strparse_i.inc - string to integer parsing functions
 *
 * (c) firk 2015, 2016, 2019, 2020, 2021, 2023; firk@cantconnect.ru
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#define STR_TO_I(ftname, uname, sname) \
extern int str_to_u##ftname(char const * st, uname * res) {\
  uname r, t;\
  char c;\
  r = 0;\
  if(!st) return -1;\
  if(!*st) return -1;\
  while(c=*st) {\
    if(c<'0' || c>'9') return -1;\
    t = r*10;\
    if(t/10!=r) return -1;\
    r = t + (unsigned char)(c-'0');\
    if(r<t) return -1;\
    st++;\
  }\
  *res = r;\
  return 0;\
} \
extern int strp_to_u##ftname(char const * * st, uname * res) {\
  char const * s;\
  uname r, t;\
  char c;\
  r = 0;\
  if(!st) return -1;\
  if(!(s=*st)) return -1;\
  while(c=*s) {\
    if(c<'0' || c>'9') break;\
    t = r*10;\
    if(t/10!=r) return -1;\
    r = t + (unsigned char)(c-'0');\
    if(r<t) return -1;\
    s++;\
  }\
  if(s==*st) return -1;\
  *st = s;\
  *res = r;\
  return 0;\
}\
extern int str_hex_to_u##ftname(char const * st, uname * res) {\
  uname r, t, u;\
  char c;\
  r = 0;\
  if(!st) return -1;\
  if(!*st) return -1;\
  while(c=*st) {\
    if((u=chr_to_value(c))>15) return -1;\
    t = r*16;\
    if(t/16!=r) return -1;\
    r = t + u;\
    if(r<t) return -1;\
    st++;\
  }\
  *res = r;\
  return 0;\
}\
extern int str_base_to_u##ftname(char const * st, uname * res, uchar b) {\
  uname r, t, u;\
  char c;\
  r = 0;\
  if(!st) return -1;\
  if(!*st) return -1;\
  while(c=*st) {\
    if((u=chr_to_value(c))>=b) return -1;\
    t = r*b;\
    if(t/b!=r) return -1;\
    r = t + u;\
    if(r<t) return -1;\
    st++;\
  }\
  *res = r;\
  return 0;\
}\
extern int strp_base_to_u##ftname(char const * * st, uname * res, uchar b) {\
  char const * s;\
  uname r, t, u;\
  char c;\
  r = 0;\
  if(!st) return -1;\
  if(!(s=*st)) return -1;\
  while(c=*s) {\
    if((u=chr_to_value(c))>=b) break;\
    t = r*b;\
    if(t/b!=r) return -1;\
    r = t + u;\
    if(r<t) return -1;\
    s++;\
  }\
  if(s==*st) return -1;\
  *st = s;\
  *res = r;\
  return 0;\
}\
extern int str_to_s##ftname(char const * st, sname * res) {\
  uname r;\
  sname rr;\
  int minus;\
  if(!st || !*st) return -1;\
  if(*st=='-') { minus=1; st++; } else\
  if(*st=='+') { minus=0; st++; } else minus=0;\
  if(str_to_u##ftname(st, &r)<0) return -1;\
  if(!r) { *res=0; return 0; }\
  if(!minus) { \
    rr = (sname)r;\
    if(rr<=0 || ((uname)rr)!=r) return -1;\
  } else { \
    rr = ((sname)-1)-(sname)(r-1);\
    if(rr>=0 || ((uname)(-1-rr))!=r-1) return -1; \
  } \
  *res = rr;\
  return 0;\
}\
extern int strp_to_s##ftname(char const * * st, sname * res) {\
  char const * s;\
  uname r;\
  sname rr;\
  int minus;\
  if(!st || !(s=*st)) return -1;\
  if(*s=='-') { minus=1; s++; } else\
  if(*s=='+') { minus=0; s++; } else minus=0;\
  if(strp_to_u##ftname(&s, &r)<0) return -1;\
  if(!r) { *st = s; *res=0; return 0; }\
  if(!minus) { \
    rr = (sname)r;\
    if(rr<=0 || ((uname)rr)!=r) return -1;\
  } else { \
    rr = ((sname)-1)-(sname)(r-1);\
    if(rr>=0 || ((uname)(-1-rr))!=r-1) return -1; \
  } \
  *st = s;\
  *res = rr;\
  return 0;\
}\
extern int str_base_to_s##ftname(char const * st, sname * res, uchar b) {\
  uname r;\
  sname rr;\
  int minus;\
  if(!st || !*st) return -1;\
  if(*st=='-') { minus=1; st++; } else\
  if(*st=='+') { minus=0; st++; } else minus=0;\
  if(str_base_to_u##ftname(st, &r, b)<0) return -1;\
  if(!r) { *res=0; return 0; }\
  if(!minus) { \
    rr = (sname)r;\
    if(rr<=0 || ((uname)rr)!=r) return -1;\
  } else { \
    rr = ((sname)-1)-(sname)(r-1);\
    if(rr>=0 || ((uname)(-1-rr))!=r-1) return -1; \
  } \
  *res = rr;\
  return 0;\
}\
extern int strp_base_to_s##ftname(char const * * st, sname * res, uchar b) {\
  char const * s;\
  uname r;\
  sname rr;\
  int minus;\
  if(!st || !(s=*st)) return -1;\
  if(*s=='-') { minus=1; s++; } else\
  if(*s=='+') { minus=0; s++; } else minus=0;\
  if(strp_base_to_u##ftname(&s, &r, b)<0) return -1;\
  if(!r) { *st = s; *res=0; return 0; }\
  if(!minus) { \
    rr = (sname)r;\
    if(rr<=0 || ((uname)rr)!=r) return -1;\
  } else { \
    rr = ((sname)-1)-(sname)(r-1);\
    if(rr>=0 || ((uname)(-1-rr))!=r-1) return -1; \
  } \
  *st = s;\
  *res = rr;\
  return 0;\
}

